/*
 * filter.c: URL filter daemon
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Jomily Joseph <jomily.joseph@iopsys.eu>
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <regex.h>
#include <arpa/inet.h>
#include <linux/netfilter.h>
#include <syslog.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <libnetfilter_queue/linux_nfnetlink_queue.h>
#include <libubox/uloop.h>
#include <pthread.h>

#include "filter.h"
#include "filter_utils.h"
#include "ubus.h"
#include "utils.h"
#include "dns_cache.h"

struct nfq_handle *nfq_h;
struct nfq_q_handle *queue_h;
struct uloop_fd uloop_fd;

// Function checks if the URL requested is globally blacklisted or not.
// @param payload_pkt packet payload
// @param url_found indicates if url was present in global list or not
// @return PKT_ACCEPT or PKT_DROP
static int global_url_filtering(char *payload_pkt, int *url_found)
{
	int i = 0;
	regex_t regex;
	char url_regex[1024];

	// Check if global blacklisting is enabled or not.
	if (true != url_global.enable_global_blacklisting) {
		syslog(LOG_INFO, "global blacklisting not enabled.");
		return PKT_ACCEPT;
	}

	// Check if the url is present in the global blacklist urls,
	// then drop the packet else accept the packet.
	for (i = 0; i < url_global.blacklist_url_cnt; i++) {
		// Compile the regular expression
		wildcard_to_regex(url_global.blacklist_url[i].url, url_regex);
		int reti = regcomp(&regex, url_regex, REG_ICASE | REG_EXTENDED);
		if (reti) {
			syslog(LOG_ERR, "Could not compile regex: %s", url_regex);
			continue;
		}

		// Execute the regular expression
		reti = regexec(&regex, payload_pkt, 0, NULL, 0);
		regfree(&regex);

		if (!reti) {
			url_global.blacklist_url[i].match_cnt++;
			return PKT_DROP;
		}
	}

	*url_found = 0;
	return PKT_ACCEPT;
}

// Function checks if the URL requested is in blacklist or whitelist for a
// particular profile.
// @param payload_pkt packet payload
// @param match_profile profile store which has the requested mac address.
// @return PKT_ACCEPT or PKT DROP on success else -1.
static int match_url_from_profile(char *payload_pkt, profile *match_profile)
{
	int i;
	regex_t regex;
	int reti;
	char url_regex[1024];

	if (!match_profile) {
		syslog(LOG_INFO, "match_profile is null.");
		return CHECK_GLOBAL_URL_LIST;
	}

	// Now check if the URL of the received packet is present in the whitelist url
	// of the above stored profile, if yes then accept the packet.
	for (i = 0; i < match_profile->whitelist_url_cnt; i++) {
		// Compile the regular expression
		wildcard_to_regex(match_profile->whitelist_url[i].url, url_regex);
		reti = regcomp(&regex, url_regex, REG_ICASE | REG_EXTENDED);
		if (reti) {
			syslog(LOG_ERR, "Could not compile regex: %s", url_regex);
			continue;
		}

		// Execute the regular expression
		reti = regexec(&regex, payload_pkt, 0, NULL, 0);
		regfree(&regex);

		if (!reti) {
			match_profile->whitelist_url[i].match_cnt++;
			return PKT_ACCEPT;
		}
	}

	// Check if the URL of the received packet is present in the blacklist url of
	// the above stored profile, if yes then drop the packet.
	for (i = 0; i < match_profile->blacklist_url_cnt; i++) {
		// Compile the regular expression
		wildcard_to_regex(match_profile->blacklist_url[i].url, url_regex);
		reti = regcomp(&regex, url_regex, REG_ICASE | REG_EXTENDED);
		if (reti) {
			syslog(LOG_ERR, "Could not compile regex: %s", url_regex);
			continue;
		}

		// Execute the regular expression
		reti = regexec(&regex, payload_pkt, 0, NULL, 0);
		regfree(&regex);

		if (!reti) {
			match_profile->blacklist_url[i].match_cnt++;
			return PKT_DROP;
		}
	}

	return CHECK_GLOBAL_URL_LIST;
}


// Function checks if the current day and time is within
// the given constrainst or not.
// @param i index of url_filter structure.
// @param day_index current day index.
// @param prev_day_index previous day index
// @param time_data structure storing data.
// @return 0 on success else -1.
static int check_if_time_within_range(int i, unsigned int day_index,
		unsigned int prev_day_index, time_info time_data)
{
	int j = 0;
	bool is_day_present = 0;

	// Check if the day is specified in the configuration.
	// If day is not specified in the filter config then the filter
	// will be applied for all days for the specified time and
	// duration.
	for (j = 0; j < MAX_DAYS; j++) {
		if (FILTER_ACTIVE == (url_filter + i)->day[j]) {
			is_day_present = 1;
			break;
		}
	}

	if (is_day_present == 0) {
		if ((time_data.current_time >= time_data.start_time) &&
				(time_data.current_time <= time_data.end_time))
			return 0;

		if (time_data.is_next_day == true) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}

		return CHECK_GLOBAL_URL_LIST;
	}

	// Now check if the filter is active for the day if yes
	// then check if is_next_day_slot is enabled and the filter
	// was active for previous day. if yes then check the
	// time range in both current day slot and next day slot.
	// Else check only in current_day_time_slot.
	if ((url_filter + i)->day[day_index] == FILTER_ACTIVE) {
		if ((time_data.current_time >= time_data.start_time) &&
				(time_data.current_time <= time_data.end_time))
			return 0;

		if (time_data.is_next_day == true &&
				(url_filter + i)->day[prev_day_index] == FILTER_ACTIVE) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}
	} else {
		// Now if the filter is not active for the day check if
		// is_next_day is enabled and the filter was active
		// for yesterday. If yes, then check the time slot from
		// next_day_time_slot.
		if (time_data.is_next_day == true &&
				(url_filter + i)->day[prev_day_index] == FILTER_ACTIVE) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}
	}

	return CHECK_GLOBAL_URL_LIST;
}

// Function filters the packet based on time slots.
// @param i index of url_filter structure.
// @return 0 on success else -1.
static int time_based_url_filtering(int i)
{
	time_info time_data;
	unsigned int day_index = 0, prev_day_index = 0,
		     duration = 0, hh = 0, ss = 0, mm = 0;

	memset(&time_data, 0, sizeof(time_info));
	// Fetch the current day and time.
	time_data.current_time = get_current_day_and_time(&day_index);

	// Get previous day index
	prev_day_index = day_index ? day_index - 1 : MAX_DAYS;

	// Check if duration is greater than 24 hrs ie 86400 seconds if yes then
	// set duration to 24 hrs.
	duration = (url_filter + i)->duration;
	if (duration >= MAX_SEC)
		duration = MAX_SEC;

	// Convert start_time into seconds and add duration to it to get the end time.
	// If end time is greater than 86400 then enable is_next_day flag and calculate the
	// next day end time.
	sscanf((url_filter + i)->start_time, "%u:%u:%u", &hh, &mm, &ss);
	time_data.start_time = ((hh * 60) + mm) * 60 + ss;
	time_data.end_time = time_data.start_time + duration;
	if (time_data.end_time > MAX_SEC) {
		time_data.is_next_day = 1;
		time_data.next_day_end_time = time_data.end_time - MAX_SEC;
		time_data.end_time = MAX_SEC;
	}

	return check_if_time_within_range(i, day_index, prev_day_index, time_data);
}



// Function checks if the requested URL is present in the profile.
// If yes, then either accepts it or drops it as per the configuration.
// @param payload_pkt packet data
// @param match_profile profile which matches with the source mac address
// @param is_https_packet https packet or not
// @return on success PKT_ACCEPT or PKT_DROP else -1.
static int mac_based_url_filtering(char *payload_pkt, profile *match_profile,
						bool is_https_packet, bool is_ipv6)
{
	int ret = CHECK_GLOBAL_URL_LIST;
	char *url_pkt = NULL;

	// If the packet is https then find the URL corresponding to
	// the IP address present in the payload_pkt from the dns cache.
	if (is_https_packet == true) {
		url_pkt = match_ip_address_from_dns_cache(payload_pkt, is_ipv6);
		if (url_pkt == NULL)
			return PKT_ACCEPT;
		ret = match_url_from_profile(url_pkt, match_profile);
	} else {
		ret = match_url_from_profile(payload_pkt, match_profile);
		if (ret == CHECK_GLOBAL_URL_LIST) {
			// Check if the url is present in cname list of dns cache.
			// If yes, then find the url name corresponding to the cname and check
			// if that is present in the matched profile. If yes, then drop the packet.
			char *cname = match_cname_from_dns_cache(payload_pkt);
			if (cname != NULL)
				ret = match_url_from_profile(cname, match_profile);
		}
	}

	return ret;
}

// Function finds the profile associated with the MAC address of the
// packet received and maps the URL with the requested URL.
// @param mac_addr mac address to be searched in the profile
// @param len length of mac_addr
// @return on success PKT_ACEEPT or PKT_DROP else -1.
static int profile_based_url_filtering(char *payload_pkt, char *mac_addr,
						size_t len, bool is_https_packet, bool is_ipv6)
{
	int i = 0, j = 0, ret = CHECK_GLOBAL_URL_LIST;
	profile *match_profile;

	// First check if this mac address is associated with any profile
	// in the filter configuration.
	for (i = 0; i < filter_cnt; i++) {
		// Check if the profile is enabled or not.
		if (((url_filter + i)->enable) == false)
			continue;

		// If mac address is found, we will check if the URL of the packet
		// received matches the one in the profile. Here we are assuming that
		// mac addresses are not duplicated in filters, if so, we use the
		// first profile in which we find the mac address.
		for (j = 0; j < (url_filter + i)->mac_cnt; j++) {
			if (strncasecmp(mac_addr, (url_filter + i)->mac_list[j], len) == 0) {
				match_profile = (url_filter + i)->p;
				ret = mac_based_url_filtering(payload_pkt,
						match_profile, is_https_packet, is_ipv6);

				// If the return value is -1 or time based url filtering
				// is not enabled then return the value from the
				// previous function.
				if (ret == CHECK_GLOBAL_URL_LIST ||
						(url_filter + i)->start_time[0] == '\0' ||
						(url_filter + i)->duration == 0) {
					return ret;
				}

				// If the packet falls between the time slot then
				// return either PKT_ACCEPT or PKT_DROP based on the
				// previous function else return -1.
				if (!time_based_url_filtering(i))
					return ret;
				else
					return CHECK_GLOBAL_URL_LIST;
			}
		}
	}

	return ret;
}

// Checks if the packet needs to be accepted or not based on mac address
// and url.
// @param payload packet payload
// @return PKT_ACCEPT or PKT_DROP
static int handle_packet_filtering(struct nfq_data *payload)
{
	struct nfqnl_msg_packet_hw *hw_hdr;
	unsigned char *data = NULL;
	char *payload_pkt = NULL;
	uint8_t proto;
	size_t data_len;
	int ret = 0;
	bool is_https_packet = false;
	bool is_ipv6 = false;

	// Get the source mac address.
	hw_hdr = nfq_get_packet_hw(payload);
	if (!hw_hdr) {
		syslog(LOG_ERR, "Error in fetching mac");
		return PKT_ACCEPT;
	}

	char mac_addr[MAC_LEN];

	memset(mac_addr, '\0', MAC_LEN);
	snprintf(mac_addr, sizeof(mac_addr), "%02x:%02x:%02x:%02x:%02x:%02x",
			hw_hdr->hw_addr[0], hw_hdr->hw_addr[1], hw_hdr->hw_addr[2],
			hw_hdr->hw_addr[3], hw_hdr->hw_addr[4], hw_hdr->hw_addr[5]);

	if (mac_addr[0] == '\0') {
		syslog(LOG_ERR, "MAC address empty");
		return PKT_ACCEPT;
	}

	ret = nfq_get_payload(payload, &data);
	if (ret <= 0) {
		syslog(LOG_ERR, "No payload found");
		return PKT_ACCEPT;
	}

	data_len = (size_t)ret;

	// Get the payload offset.
	int payload_offset;

	payload_offset = get_ip_packet_payload(data, &proto, data_len, &is_ipv6);
	if (payload_offset <= 0) {
		// Just accept the packet and return since the packet either does not math
		// urlfilter criteria or is a dns response.
		return PKT_ACCEPT;
	}

	ret = get_payload_pkt(proto, data, payload_offset,
			&payload_pkt, &is_https_packet, is_ipv6);

	if (ret != URL_SUCCESS)
		return ret;

	if (payload_pkt == NULL) {
		syslog(LOG_ERR, "payload pkt NULL");
		return PKT_ACCEPT;
	}

	size_t len = strlen(payload_pkt);

	if (len == 0) {
		syslog(LOG_ERR, "payload is empty");
		return PKT_ACCEPT;
	}

	// Check if the URL has been blacklisted in any profile.
	ret = profile_based_url_filtering(payload_pkt, mac_addr, sizeof(mac_addr), is_https_packet, is_ipv6);
	if (ret == CHECK_GLOBAL_URL_LIST) {
		// The destination (url or ip) is not blacklisted in a profile, check if the URL has been
		// globally blacklisted or not. If yes then drop the packet else accept the packet.
		char *url_name = NULL;
		int url_found = 1;

		if (is_https_packet == true) {

			url_name = match_ip_address_from_dns_cache(payload_pkt, is_ipv6);

			if (url_name == NULL) {
				// destination ip does not belong to a blacklisted url
				return PKT_ACCEPT;
			} else {
				// url found in dns cache, check if this url is globally blacklisted
				return global_url_filtering(url_name, &url_found);
			}
		} else {
			// Likely DNS packet, check url for blacklisting
			ret = global_url_filtering(payload_pkt, &url_found);
			if (ret == PKT_DROP)
				return ret; // no need to process further

			if (url_found == 0) {
				// If yes, then find the url name corresponding to the cname and check
				// if that is present in the global list. If yes, then drop the packet.
				char *cname = match_cname_from_dns_cache(payload_pkt);
				if (cname  != NULL)
					ret = global_url_filtering(cname, &url_found);
			}
		}
	}

	return ret;
}


// Callback function for handling packets.
// @param queue_h queue 0 handle
// @param nfmsg nfq message received
// @param nfa message data
// @param data store data
// @return 0 on success and non zero otherwise
static int handle_packets(struct nfq_q_handle *q_h, struct nfgenmsg *nfmsg,
		struct nfq_data *nfa, void *data)
{
	struct nfqnl_msg_packet_hdr *pkt_hdr;
	int pkt_decision;
	uint32_t pkt_id = 0;

	// Fetch the packet header.
	pkt_hdr = nfq_get_msg_packet_hdr(nfa);
	if (pkt_hdr)
		pkt_id = ntohl(pkt_hdr->packet_id);

	// Check if we should block this packet on the basis of URL and source mac address.
	pkt_decision = handle_packet_filtering(nfa);
	if (pkt_decision == PKT_ACCEPT)
		return nfq_set_verdict(q_h, pkt_id, NF_ACCEPT, 0, NULL);
	else
		return nfq_set_verdict(q_h, pkt_id, NF_DROP, 0, NULL);
}

// Function to unbind and bind to nf_queue handler.
// @return nothing
static void update_nf_queue_handler(void)
{
	int nfq_bind_v4 = 1, nfq_bind_v6 = 1;

	// Unbinding existing nf_queue handler.
	if (nfq_unbind_pf(nfq_h, AF_INET) < 0) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf() for IPv4");
		nfq_bind_v4 = 0;
	}

	if (nfq_unbind_pf(nfq_h, AF_INET6) < 0) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf() for IPv6");
		nfq_bind_v6 = 0;
	}

	if (!(nfq_bind_v4 || nfq_bind_v6)) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf()");
		exit(1);
	}

	// Binding nfnetlink_queue as nf_queue handler.
	nfq_bind_v4 = nfq_bind_v6 = 1;
	if (nfq_bind_pf(nfq_h, AF_INET < 0)) {
		syslog(LOG_ERR, "Error during nfq_bind_pf() for IPv4");
		nfq_bind_v4 = 0;
	}

	if (nfq_bind_pf(nfq_h, AF_INET6) < 0) {
		syslog(LOG_ERR, "Error during nfq_bind_pf() for IPv6");
		nfq_bind_v6 = 0;
	}

	if (!(nfq_bind_v4 || nfq_bind_v6)) {
		syslog(LOG_ERR, "Error during nfq_bind_pf()");
		exit(1);
	}
}

// Open a netlink connection and returns file descriptor
// @return fd of the netlink socket
static int open_netlink_connection(void)
{
	// Opening netlink socket.
	nfq_h = nfq_open();
	if (!nfq_h) {
		syslog(LOG_ERR, "Error during nfq_open()");
		exit(1);
	}

	update_nf_queue_handler();

	// Binding this socket to queue.
	queue_h = nfq_create_queue(nfq_h,  0, &handle_packets, NULL);
	if (!queue_h) {
		syslog(LOG_ERR, "Error during nfq_create_queue()");
		exit(1);
	}

	// Setting copy_packet mode.
	if (nfq_set_mode(queue_h, NFQNL_COPY_PACKET, 0xffff) < 0) {
		syslog(LOG_ERR, "nfq_set_mode error");
		exit(1);
	}

	return nfq_fd(nfq_h);
}

// This function prints the usage of the daemon
static void usage(char *prog)
{
	fprintf(stderr, "Usage: %s [options]\n", prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "    -s <socket path>   ubus socket\n");
	fprintf(stderr, "    -h <help>       To print this help menu\n");
	fprintf(stderr, "\n");
}

// This function reads the packets received on the fd
// @param fd is the pointer to the file descriptor to read on
static void url_filter_recv(struct uloop_fd *fd, unsigned int events)
{
	char buf[BUFSIZE];

	(void)events;

	memset(buf, 0, sizeof(buf));
	int rv = (int)recv(fd->fd, buf, sizeof(buf), 0);
	if (rv >= 0) {
		nfq_handle_packet(nfq_h, buf, rv);
	}
}

int main(int argc, char **argv)
{
	const char *socket = NULL;
	int ch;

	while ((ch = getopt(argc, argv, "hs:")) != -1) {
		switch (ch) {
		case 's':
			socket = optarg;
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
		default:
			break;
				}
	}

	openlog("urlfilter", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	if (urlfilter_ubus_init(socket) < 0) {
		fprintf(stderr, "Failed to connect to ubus\n");
		goto exit;
	}

	// Call function to update the global structures.
	if (read_config()) {
		syslog(LOG_ERR, "failed to read config");
		goto exit;
	}

	// Create a thread which would send out dns query
	// for all the blacklisted urls to create a dns cache.
	pthread_t ptid;

	if (pthread_create(&ptid, NULL, &send_dns_query, NULL) != 0) {
		syslog(LOG_ERR, "pthread create failed");
		goto exit;
	}

	uloop_fd.fd = open_netlink_connection();
	uloop_fd.cb = url_filter_recv;
	if (uloop_fd_add(&uloop_fd, ULOOP_READ)) {
		syslog(LOG_ERR, "url fitler: uloop_fd_add(%d) failed", uloop_fd.fd);
		goto destroy;
	}

	// Main loop of urlfilter
	uloop_run();

destroy:
	// Destroying the queue handle and closing nfq handle.
	nfq_destroy_queue(queue_h);
	nfq_close(nfq_h);

exit:
	urlfilter_ubus_done();
	uloop_done();

	return 0;
}
