/*
 * filter_utils.c: URL filter utils
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <stdbool.h>
#include <syslog.h>

#include "dns.h"
#include "config.h"
#include "utils.h"
#include "filter_utils.h"

static size_t ipv6_header_get_L4_header_offset(const struct ip6_hdr *ip6h, uint8_t *nexthdr_p)
{
	unsigned int ext_head_count = 8;
	const struct ip6_ext *ip_ext_p = (const struct ip6_ext *)(ip6h + 1);
	unsigned int payload_offset = sizeof(struct ip6_hdr);

	*nexthdr_p = ip6h->ip6_nxt;
	do {
		if ((*nexthdr_p == IPPROTO_TCP) || (*nexthdr_p == IPPROTO_UDP))
			break;

		payload_offset += (ip_ext_p->ip6e_len + 1U) << 3U;
		*nexthdr_p = ip_ext_p->ip6e_nxt;
		ip_ext_p = (struct ip6_ext *)((uint8_t *)ip6h + payload_offset);
		ext_head_count--; /* at most 8 extension headers */
	} while (ext_head_count);

	return payload_offset;
}

// Check if the packet is https or not. If yes then,
// returns the destination IP address.
// @param data payload data
// @return On success destination ip address else NULL
static char *check_https_packet(unsigned char *data, char **payload_pkt, bool is_ipv6)
{
	struct sockaddr_in dest;
	const struct iphdr *iph;
	const struct tcphdr *tcp_header;
	uint8_t proto;
	const struct ip6_hdr *ip6h;
	struct sockaddr_in6 dest6;

	if (!is_ipv6) {
		iph = (const struct iphdr *)data;
		tcp_header = (const struct tcphdr *)(data + (iph->ihl<<2));

		memset(&dest, 0, sizeof(dest));
		dest.sin_addr.s_addr = iph->daddr;

		// Check if destination port is 443.
		if (ntohs(tcp_header->dest) == HTTPS_PORT)
			return inet_ntoa(dest.sin_addr);
	} else {
		ip6h = (const struct ip6_hdr *)data;
		tcp_header = (const struct tcphdr *)(data + ipv6_header_get_L4_header_offset((struct ip6_hdr *)data, &proto));

		memset(&dest6, 0, sizeof(dest6));
		memcpy(&dest6.sin6_addr, &ip6h->ip6_dst, sizeof(struct in6_addr));

		if (ntohs(tcp_header->dest) == HTTPS_PORT)
			return (char *)inet_ntop(AF_INET6, (const struct sockaddr_in6 *)&dest6.sin6_addr, *payload_pkt, MAX_IPV6_LEN);
	}

	return NULL;
}

static int handle_udp_packet(const unsigned char *data_p, size_t offset, size_t len)
{
	const struct udphdr *udp_header = (const struct udphdr *)(data_p + offset);

	if (((ntohs(udp_header->dest)) != DNS_PORT) && ((ntohs(udp_header->source)) != DNS_PORT)) {
		syslog(LOG_ERR, "Not a DNS packet");
		return -1;
	}

	if (ntohs(udp_header->source) == DNS_PORT) {
		unsigned char dns_payload[len + 1];

		memset(dns_payload, '\0', sizeof(dns_payload));
		memcpy(dns_payload, data_p + offset + sizeof(struct udphdr), len - (offset + sizeof(struct udphdr)) + 1);
		if (dns_payload[0] == '\0') {
			syslog(LOG_ERR, "payload is null");
			return -1;
		}
		parse_dns_response(dns_payload);
		return 0;
	}
	return (int)(offset + sizeof(struct udphdr) + sizeof(struct dnshdr));
}

static int handle_tcp_packet(const unsigned char *data_p, size_t offset, size_t len)
{
	const struct tcphdr *tcp_header = (const struct tcphdr *)(data_p + offset);
	size_t tcp_header_length = tcp_header->doff * 4;

	if (ntohs(tcp_header->source) == DNS_PORT) {
		unsigned char dns_payload[len + 1];

		memset(dns_payload, '\0', sizeof(dns_payload));
		if (len <= (offset + tcp_header_length)) {
			return -1;
		}

		memcpy(dns_payload, data_p + offset + tcp_header_length, len - (offset + tcp_header_length) + 1);
		if (dns_payload[0] == '\0') {
			syslog(LOG_ERR, "payload is null");
			return -1;
		}

		parse_dns_response(dns_payload);
		return 0;
	} else if (ntohs(tcp_header->dest) == DNS_PORT) {
		return (int)(offset + tcp_header_length + sizeof(struct dnshdr));
	}

	return (int)(offset + tcp_header_length);
}

static int process_ipv4_packet(const unsigned char *data_p, uint8_t *proto, size_t len, bool *is_ipv6) {
	const struct iphdr *iph = (const struct iphdr *)data_p;
	size_t ip_header_length = iph->ihl * 4;
	*proto = iph->protocol;
	*is_ipv6 = false;

	if (*proto == IPPROTO_UDP) {
		return handle_udp_packet(data_p, ip_header_length, len);
	} else if (*proto == IPPROTO_TCP) {
		return handle_tcp_packet(data_p, ip_header_length, len);
	} else {
		syslog(LOG_ERR, "Unknown protocol");
		return -1;
	}
}

static int process_ipv6_packet(const unsigned char *data_p, uint8_t *proto, size_t len, bool *is_ipv6)
{
	const struct ip6_hdr *ip6h = (const struct ip6_hdr *)data_p;

	if ((ip6h->ip6_vfc >> 4) != 6) {
		return -1;
	}

	*is_ipv6 = true;

	size_t payload_offset = ipv6_header_get_L4_header_offset(ip6h, proto);
	if (*proto == IPPROTO_UDP) {
		return handle_udp_packet(data_p, payload_offset, len);
	} else if (*proto == IPPROTO_TCP) {
		return handle_tcp_packet(data_p, payload_offset, len);
	} else {
		syslog(LOG_ERR, "Unknown protocol");
		return -1;
	}
}

int get_ip_packet_payload(unsigned char *data_p, uint8_t *proto, size_t len, bool *is_ipv6)
{
	const struct iphdr *iph = (const struct iphdr *)data_p;
	bool ipv4 = (iph->version == 4);

	if (ipv4) {
		return process_ipv4_packet(data_p, proto, len, is_ipv6);
	} else {
		return process_ipv6_packet(data_p, proto, len, is_ipv6);
	}
}

// Function fetches the packet payload ie URL for http packet,
// destination IP address for https packet and domain name for
// dns packet
// @param proto protocol type
// @param data payload
// @param payload_offset data offset
// @param payload_pkt variable to store payload
// @param is_https_packet variable to store if its a https packet
// @return 1 or PKT_ACCEPT
int get_payload_pkt(int proto, unsigned char *data, int payload_offset,
		char **payload_pkt, bool *is_https_packet, bool is_ipv6)
{
	if (proto == IPPROTO_TCP) {
		// If the packet is a dns reponse then accept the packet.
		if (payload_offset == 0)
			return PKT_ACCEPT;
		// Check if its https packet or not, if yes then update the
		// payload_pkt with destination ip address.
		*payload_pkt = check_https_packet(data, payload_pkt, is_ipv6);

		if (*payload_pkt != NULL && *payload_pkt[0] != '\0') {
			*is_https_packet = true;
			return URL_SUCCESS;
		}

		// Since its not an https packet, check if it is an http packet or not
		*payload_pkt = (char *)(data + payload_offset);
		if (strstr(*payload_pkt, "GET ") == NULL  &&
				strstr(*payload_pkt, "POST ") == NULL &&
				strstr(*payload_pkt, "HEAD ") == NULL) {
			syslog(LOG_INFO, "No http/https packet");
			return PKT_ACCEPT;
		}
	} else if (proto == IPPROTO_UDP) {
		// If the packet is a dns reponse then accept the packet.
		if (payload_offset == 0)
			return PKT_ACCEPT;

		// Get the qname(hostname) from the packet.
		*payload_pkt = get_domain_name((char *)(data + payload_offset));
		if (*payload_pkt == NULL) {
			syslog(LOG_INFO, "Payload for dns packet is NULL");
			return PKT_ACCEPT;
		}
	}

	return URL_SUCCESS;
}
