/*
 * utils.c: URL filter misc. utils
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <time.h>

#include "utils.h"
/* return 1 if IP string is valid, else return 0 */
int validate_ip(char *ip_str)
{
	struct sockaddr_in sa;

	return inet_pton(AF_INET, ip_str, &(sa.sin_addr));

}

/* return 1 if IPv6 string is valid, else return 0 */
int validate_ipv6(char *ip6_str)
{
	struct sockaddr_in6 sa6;

	return inet_pton(AF_INET6, ip6_str, &(sa6.sin6_addr));
}

// Function to convert wildcard pattern to regex pattern
void wildcard_to_regex(const char *wildcard, char *regex)
{
	char *p = regex;
	*p++ = '^'; // Match the start of the string

	while (*wildcard) {
		if (*wildcard == '*') {
			*p++ = '.';
			*p++ = '*';
		} else if (*wildcard == '.') {
			*p++ = '\\';
			*p++ = '.';
		} else {
			*p++ = *wildcard;
		}
		wildcard++;
	}

	*p++ = '$'; // Match the end of the string
	*p = '\0';

	return;
}

// Fetches the current day and time
// @param day_index index of the day of week
// @return current time in seconds.
unsigned int get_current_day_and_time(unsigned int *day_index)
{
	const char *weekdays[] = {"Sunday",
				"Monday", "Tuesday",
				"Wednesday", "Thursday",
				"Friday", "Saturday"};
	char day[MAX_LEN_WEEKDAY];
	time_t cur_time;
	struct tm *loc_time;
	unsigned int i = 0;

	// First fetch the day
	memset(day, '\0', MAX_LEN_WEEKDAY);
	cur_time = time(NULL);
	loc_time = localtime(&cur_time);
	strftime(day, MAX_LEN_WEEKDAY, "%A", loc_time);
	day[strcspn(day, "\r\n")] = 0;

	// Fetch index for the day.
	for (i = 0; i < MAX_DAYS; i++) {
		if (strncasecmp(day, weekdays[i], sizeof(day)) == 0) {
			*day_index = i;
			break;
		}
	}

	// Get the current time in seconds format.
	return (unsigned int)(((loc_time->tm_hour * 60) + loc_time->tm_min) * 60 + loc_time->tm_sec);
}

// Function gets the qname ie the hostname for which the dns query
// has been generated, from the packet
// @param data payload
// @domaaain_name variable to store the url
// @return nothing
char *get_domain_name(char data[])
{
	// The content of the QNAME is of the format 3www8mydomain3com0
	// instead of www.mydomain.com. Also this is a variable length
	// field to fit the hostname. So we need to parse the data to
	// get the url for which the dns query has been initiated.
	int i = 0, j = 0, start = 0;
	int toread = data[0];

	i++;

	while (toread != 0) {
		for (; i <= start+toread; i++) {
			data[j] = data[i];
			j++;
		}

		// Get next chunk
		data[j] = '.';
		j++;
		toread = data[i];
		start = i;
		i++;
	}

	data[i - 2] = '\0';
	return data;
}
